# -*- coding: utf-8 -*-

"""Collection of additional helper utilities for dialog management."""

__all__ = ['WidgetGroup']

import typing
from qtpy import QtWidgets

if typing.TYPE_CHECKING:
    import typing_extensions


class WidgetGroup(typing.Sequence[QtWidgets.QWidget]):
    """Collection of widgets, with ability for synchronous control."""

    __slots__ = ('__content',)

    def __init__(self, *args: typing.Union[QtWidgets.QWidget, typing.Iterable[QtWidgets.QWidget]]) -> None:
        """Initialize new :class:`~WidgetGroup` instance."""
        self.__content: 'typing_extensions.Final[typing.List[QtWidgets.QWidget]]' = []

        arg: typing.Union[QtWidgets.QWidget, typing.Iterable[QtWidgets.QWidget]]
        for arg in args:
            if isinstance(arg, QtWidgets.QWidget):
                self.__content.append(arg)
            else:
                self.__content.extend(arg)

    def disable(self, state: bool = True) -> None:
        """Set disabled state to all widgets in the group."""
        self.for_each(action=lambda item: item.setDisabled(state))

    def enable(self, state: bool = True) -> None:
        """Set enabled state to all widgets in the group."""
        self.for_each(action=lambda item: item.setEnabled(state))

    def for_each(self, action: typing.Callable[[QtWidgets.QWidget], typing.Any]) -> None:
        """Preform same `action` for each widget in the group."""
        item: 'QtWidgets.QWidget'
        for item in self.__content:
            action(item)

    def hide(self, state: bool = True) -> None:
        """Set hidden state to all widgets in the group."""
        self.for_each(action=lambda item: item.setHidden(state))

    def only(self, predicate: typing.Callable[[QtWidgets.QWidget], bool]) -> 'WidgetGroup':
        """Filter widgets in group by predicate."""
        return WidgetGroup(
            item
            for item in self.__content
            if predicate(item)
        )

    def only_disabled(self) -> 'WidgetGroup':
        """Select widgets that are disabled."""
        return self.only(predicate=lambda item: not item.isEnabled())

    def only_enabled(self) -> 'WidgetGroup':
        """Select widgets that are enabled."""
        return self.only(predicate=lambda item: item.isEnabled())

    def only_hidden(self) -> 'WidgetGroup':
        """Select widgets that are hidden."""
        return self.only(predicate=lambda item: not item.isVisible())

    def only_visible(self) -> 'WidgetGroup':
        """Select widgets that are visible."""
        return self.only(predicate=lambda item: item.isVisible())

    def show(self, state: bool = True) -> None:
        """Set visible state to all widgets in cgroup."""
        self.for_each(action=lambda item: item.setVisible(state))

    def __add__(self, other: 'WidgetGroup') -> 'WidgetGroup':
        """Concatenate two widget groups."""
        if not isinstance(other, WidgetGroup):
            return NotImplemented  # type: ignore
        return WidgetGroup(self, other)

    @typing.overload
    def __getitem__(self, index: int) -> 'QtWidgets.QWidget':
        """Retrieve single widget by its index."""

    @typing.overload
    def __getitem__(self, index: slice) -> 'WidgetGroup':
        """Retrieve subset of widgets."""

    def __getitem__(self, index: typing.Union[int, slice]) -> typing.Union['QtWidgets.QWidget', 'WidgetGroup']:
        """Retrieve content of the collection."""
        if isinstance(index, int):
            return self.__content[index]
        return WidgetGroup(self.__content[index])

    def __len__(self) -> int:
        """Retrieve total number of items in the collection."""
        return len(self.__content)



# -*- coding: utf-8 -*-

# pylint: disable=too-few-public-methods

"""Collection of application detection utilities."""

__all__ = [
    'join',
    'Linux', 'Mac', 'Win',
    'DetectedApplication', 'Group',
    'CheckConfiguredRoots', 'CheckKnownRoots', 'check_known_mac_roots', 'AppendExecutable',
]

import inspect
import itertools
import os
import typing
from anaconda_navigator import config as navigator_config

if typing.TYPE_CHECKING:
    import typing_extensions
    from anaconda_navigator.config import user as user_config

    class Source(typing_extensions.Protocol):
        """
        Initial source of application details.

        It is not required for application details to be complete - they still might be extended with filters.
        """

        def __call__(self) -> typing.Iterator['DetectedApplication']:
            """Iterate through detected applications."""

    class Filter(typing_extensions.Protocol):
        """
        Filters for application details.

        Filters are allowed to generate new application details, as well as modify existing ones before forwarding them
        further.
        """

        def __call__(self, parent: typing.Iterator['DetectedApplication']) -> typing.Iterator['DetectedApplication']:
            """Iterate through detected applications."""


def is_source(
        instance: typing.Callable[..., typing.Iterator['DetectedApplication']],
) -> 'typing_extensions.TypeGuard[Source]':
    """
    Check if `instance` is :class:`~Source` instance.

    This check is a rough one, as we don't have a strict type limitations.
    """
    signature: inspect.Signature = inspect.signature(instance)
    return len(signature.parameters) == 0


def is_filter(
        instance: typing.Callable[..., typing.Iterator['DetectedApplication']],
) -> 'typing_extensions.TypeGuard[Filter]':
    """
    Check if `instance` is :class:`~Filter` instance.

    This check is a rough one, as we don't have a strict type limitations.
    """
    signature: inspect.Signature = inspect.signature(instance)
    if len(signature.parameters) != 1:
        return False
    parent: typing.Optional[inspect.Parameter] = signature.parameters.get('parent', None)
    return (parent is not None) and (parent.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD)


def join(root: typing.Optional[str], *args: typing.Optional[str]) -> typing.Optional[str]:
    """Join path parts, if all of them are not :code:`None`."""
    if root is None:
        return None

    if any(arg is None for arg in args):
        return None
    args = typing.cast(typing.Tuple[str, ...], args)

    return os.path.abspath(os.path.join(root, *args))


class Linux:
    """Collection of Linux locations."""

    def __init__(self) -> None:
        """Initialize new :class:`~Linux` instance."""
        raise NotImplementedError()

    root: typing.ClassVar[typing.Optional[str]] = None
    home: typing.ClassVar[typing.Optional[str]] = None

    if navigator_config.LINUX:
        root = '/'
        home = os.path.expanduser('~')


class Mac:
    """Collection of OS X locations."""

    def __init__(self) -> None:
        """Initialize new :class:`~Mac` instance."""
        raise NotImplementedError()

    root: typing.ClassVar[typing.Optional[str]] = None
    home: typing.ClassVar[typing.Optional[str]] = None

    applications: typing.ClassVar[typing.Optional[str]] = None
    user_applications: typing.ClassVar[typing.Optional[str]] = None

    if navigator_config.MAC:
        root = '/'
        home = os.path.expanduser('~')

        applications = os.path.join(root, 'Applications')
        user_applications = os.path.join(home, 'Applications')


class Win:
    """Collection of Windows locations."""

    def __init__(self) -> None:
        """Initialize new :class:`~Win` instance."""
        raise NotImplementedError()

    home: typing.ClassVar[typing.Optional[str]] = None

    program_files_x86: typing.ClassVar[typing.Optional[str]] = None
    program_files_x64: typing.ClassVar[typing.Optional[str]] = None

    program_data: typing.ClassVar[typing.Optional[str]] = None

    local_app_data: typing.ClassVar[typing.Optional[str]] = None
    local_app_data_low: typing.ClassVar[typing.Optional[str]] = None
    roaming_app_data: typing.ClassVar[typing.Optional[str]] = None

    if navigator_config.WIN:
        from anaconda_navigator.external import knownfolders  # pylint: disable=import-outside-toplevel

        home = os.path.expanduser('~')

        program_files_x86 = knownfolders.get_folder_path(knownfolders.FOLDERID.ProgramFilesX86)[0]  # type: ignore
        program_files_x64 = knownfolders.get_folder_path(knownfolders.FOLDERID.ProgramFilesX64)[0]  # type: ignore

        program_data = knownfolders.get_folder_path(knownfolders.FOLDERID.ProgramData)[0]  # type: ignore

        local_app_data = knownfolders.get_folder_path(knownfolders.FOLDERID.LocalAppData)[0]  # type: ignore
        local_app_data_low = knownfolders.get_folder_path(knownfolders.FOLDERID.LocalAppDataLow)[0]  # type: ignore
        roaming_app_data = knownfolders.get_folder_path(knownfolders.FOLDERID.RoamingAppData)[0]  # type: ignore


KEEP: typing.Any = object()


class DetectedApplication:
    """Description of the detected application."""

    __slots__ = ('__root', '__executable', '__version')

    def __init__(
            self,
            *,
            root: typing.Optional[str] = None,
            executable: typing.Optional[str] = None,
            version: typing.Optional[str] = None,
    ) -> None:
        """Initialize new :class:`~DetectedApplication`."""
        self.__root: 'typing_extensions.Final[typing.Optional[str]]' = root
        self.__executable: 'typing_extensions.Final[typing.Optional[str]]' = executable
        self.__version: 'typing_extensions.Final[typing.Optional[str]]' = version

    @property
    def complete(self) -> bool:  # noqa: D401
        """All required values are provided."""
        return all(
            item is not None
            for item in (self.__root, self.__executable, self.__version)
        )

    @property
    def root(self) -> str:  # noqa: D401
        """Root directory of the detected application."""
        if self.__root is None:
            raise AttributeError('Incomplete detected application description')
        return self.__root

    @property
    def executable(self) -> str:  # noqa: D401
        """Path to the executable of the detected application."""
        if self.__executable is None:
            raise AttributeError('Incomplete detected application description')
        return self.__executable

    @property
    def version(self) -> str:  # noqa: D401
        """Version of the detected application."""
        if self.__version is None:
            raise AttributeError('Incomplete detected application description')
        return self.__version

    def replace(
            self,
            *,
            root: typing.Optional[str] = KEEP,
            executable: typing.Optional[str] = KEEP,
            version: typing.Optional[str] = KEEP,
    ) -> 'DetectedApplication':
        """Prepare a copy of current instance with some additional replacements."""
        if root is KEEP:
            root = self.__root
        if executable is KEEP:
            executable = self.__executable
        if version is KEEP:
            version = self.__version
        return DetectedApplication(root=root, executable=executable, version=version)


class Group:
    """
    Group of application detecting sources and filters.

    Order in which you provide sources and filters matters. Output of each source is chained with all previously
    detected applications (if such). Filters are applied to all previously detected applications.
    """

    __slots__ = ('__content',)

    def __init__(self, *args: typing.Union['Source', 'Filter']) -> None:
        """Initialize new :class:`~Group` instance."""
        if len(args) <= 0:
            raise TypeError('at least one source must be provided')

        if not is_source(args[0]):
            raise TypeError('first argument must be a Source')

        arg: typing.Union['Source', 'Filter']
        for arg in args:
            if (not is_source(arg)) and (not is_filter(arg)):
                raise TypeError(f'each argument must be source- of filter-compatible, not {type(arg).__name__}')

        self.__content: 'typing_extensions.Final[typing.Tuple[typing.Union[Source, Filter], ...]]' = args

    def __call__(self) -> typing.Iterator[DetectedApplication]:
        """Iterate through detected applications."""
        result: typing.Iterator[DetectedApplication] = typing.cast('Source', self.__content[0])()

        arg: typing.Union['Source', 'Filter']
        for arg in itertools.islice(self.__content, 1, len(self.__content), 1):
            if is_source(arg):
                result = itertools.chain(result, arg())
            else:
                arg = typing.cast('Filter', arg)
                result = arg(result)

        return result


def collect_str(source: typing.Union[None, str, typing.Iterable[typing.Any]]) -> typing.Iterator[str]:
    """Collect string values from the recursive collections/iterables of :class:`~str` or :code:`None` instances."""
    stack: typing.List[typing.Union[None, str, typing.Iterable[typing.Any]]] = [source]
    while stack:
        current: typing.Union[None, str, typing.Iterable[typing.Any]] = stack.pop()

        if current is None:
            continue

        if isinstance(current, str):
            yield current

        else:
            offset: int = len(stack)
            child: typing.Union[None, str, typing.Iterable[typing.Any]]
            for child in current:
                stack.insert(offset, child)


class CheckConfiguredRoots:
    """
    Get application root from the configuration.

    Multiple configuration keys might be provided as `args`.

    Configuration, from which value should be retrieved, must be provided as a keyword argument.

    If configuration section is not provided - "main" will be used by default.
    """

    __slots__ = ('__content', '__configuration', '__section')

    def __init__(
            self,
            *args: typing.Union[None, str, typing.Iterable[typing.Optional[str]]],
            configuration: 'user_config.UserConfig',
            section: str = 'main',
    ) -> None:
        """Initialize new :class:`~CheckKnownRoots` instance."""
        self.__content: 'typing_extensions.Final[typing.Tuple[str, ...]]' = tuple(collect_str(args))
        self.__configuration: 'typing_extensions.Final[user_config.UserConfig]' = configuration
        self.__section: 'typing_extensions.Final[str]' = section

    def __call__(self) -> typing.Iterator[DetectedApplication]:
        """Iterate through detected applications."""
        setting: str
        for setting in self.__content:
            root: str = self.__configuration.get(self.__section, setting, '')
            if not root:
                continue

            root = os.path.abspath(root)
            if os.path.exists(root):
                yield DetectedApplication(root=root)


class CheckKnownRoots:
    """
    Iterate through known roots, selecting only existing ones.

    Some of root options might be :code:`None` - they will be skipped.
    """

    __slots__ = ('__content',)

    def __init__(self, *args: typing.Union[None, str, typing.Iterable[typing.Optional[str]]]) -> None:
        """Initialize new :class:`~CheckKnownRoots` instance."""
        self.__content: 'typing_extensions.Final[typing.Tuple[str, ...]]' = tuple(collect_str(args))

    def __call__(self) -> typing.Iterator[DetectedApplication]:
        """Iterate through detected applications."""
        root: str
        for root in self.__content:
            root = os.path.abspath(root)
            if os.path.exists(root):
                yield DetectedApplication(root=root)


def check_known_mac_roots(*args: typing.Union[None, str, typing.Iterable[typing.Optional[str]]]) -> CheckKnownRoots:
    """
    Shortcut for checking application roots of OS X applications.

    Unlike :class:`~CheckKnownRoots` - you need to provide application names instead of directory paths.
    """
    return CheckKnownRoots(*(
        join(root, arg)
        for root in (Mac.user_applications, Mac.applications)
        for arg in collect_str(args)
    ))


class StepIntoRoot:
    """
    Iterate through children of application root.

    Each child might be checked that it `starts_with`, `ends_with` or `equals` to one or multiple values.

    Iteration results are sorted. If you want descending order of the results - set `reverse` to :code:`True`.
    """

    __slots__ = ('__equals', '__starts_with', '__ends_with', '__reverse')

    def __init__(
            self,
            *,
            equals: typing.Union[str, typing.Iterable[str]] = (),
            starts_with: typing.Union[str, typing.Iterable[str]] = (),
            ends_with: typing.Union[str, typing.Iterable[str]] = (),
            reverse: bool = False,
    ) -> None:
        """Initialize new :class:`~StepIntoRoot` instance."""
        if isinstance(equals, str):
            equals = {equals}
        else:
            equals = set(equals)

        if isinstance(starts_with, str):
            starts_with = (starts_with,)
        else:
            starts_with = tuple(starts_with)

        if isinstance(ends_with, str):
            ends_with = (ends_with,)
        else:
            ends_with = tuple(ends_with)

        self.__equals: 'typing_extensions.Final[typing.Set[str]]' = equals
        self.__starts_with: 'typing_extensions.Final[typing.Tuple[str, ...]]' = starts_with
        self.__ends_with: 'typing_extensions.Final[typing.Tuple[str, ...]]' = ends_with
        self.__reverse: 'typing_extensions.Final[bool]' = reverse

    def __call__(self, parent: typing.Iterator[DetectedApplication]) -> typing.Iterator[DetectedApplication]:
        """Iterate through detected applications."""
        application: DetectedApplication
        for application in parent:
            if not application.root:
                continue

            children: typing.List[str]
            try:
                children = sorted(os.listdir(application.root), reverse=self.__reverse)
            except OSError:
                continue

            child: str
            for child in children:
                if self.__equals and (child not in self.__equals):
                    continue
                if self.__starts_with and not any(child.startswith(item) for item in self.__starts_with):
                    continue
                if self.__ends_with and not any(child.endswith(item) for item in self.__ends_with):
                    continue
                yield application.replace(root=os.path.join(application.root, child))


class AppendExecutable:
    """
    Check multiple options of executables and iterate through existing ones.

    Each option of the executable must be a path relative to the application root.
    """

    __slots__ = ('__content',)

    def __init__(self, *args: typing.Union[None, str, typing.Iterable[typing.Optional[str]]]) -> None:
        """Initialize new :class:`~AppendExecutable` instance."""
        self.__content: 'typing_extensions.Final[typing.Tuple[str, ...]]' = tuple(collect_str(args))

    def __call__(self, parent: typing.Iterator[DetectedApplication]) -> typing.Iterator[DetectedApplication]:
        """Iterate through detected applications."""
        application: DetectedApplication
        executable: str
        for application in parent:
            for executable in self.__content:
                executable = os.path.abspath(os.path.join(application.root, executable))
                if os.path.exists(executable):
                    yield application.replace(executable=executable)


class If:
    """Iterate through application only if a `condition` is met."""

    __slots__ = ('__condition',)

    def __init__(self, condition: bool) -> None:
        """Initialize new :class:`~If` instance."""
        self.__condition: 'typing_extensions.Final[bool]' = condition

    def __call__(self, parent: typing.Iterator[DetectedApplication]) -> typing.Iterator[DetectedApplication]:
        """Iterate through detected applications."""
        if self.__condition:
            return parent

        return iter(())


def linux_only() -> If:
    """Iterate through applications only if on a linux machine."""
    return If(navigator_config.LINUX)


def mac_only() -> If:
    """Iterate through applications only if on an os x machine."""
    return If(navigator_config.MAC)


def win_only() -> If:
    """Iterate through applications only if on a windows machine."""
    return If(navigator_config.WIN)



# -*- coding: utf-8 -*-

"""Collection of the core utilities."""

from __future__ import annotations

__all__ = ['is_conda_available', 'get_conda_info']

import os
import sys
import typing
import json
from . import exceptions
from . import launch
from . import solvers

if typing.TYPE_CHECKING:
    import typing_extensions
    from . import types as conda_types


WIN: 'typing_extensions.Final[bool]' = (os.name == 'nt')


def get_conda_cmd_path() -> typing.Optional[str]:
    """Check if conda is found on path."""
    bin_folder: str = 'Scripts' if WIN else 'bin'
    conda_exe: str = 'conda-script.py' if WIN else 'conda'
    env_prefix: str = os.path.dirname(os.path.dirname(sys.prefix))
    cmds: typing.List[str] = [
        os.path.join(env_prefix, bin_folder, conda_exe),
        os.path.join(sys.prefix, bin_folder, conda_exe),
        'conda',
    ]

    cmd: str
    for cmd in cmds:
        stdout: str
        stderr: str
        error: bool
        stdout, stderr, error = launch.run_process(cmd_list=[cmd, '--version'])

        if (not error) and any(item.startswith('conda ') for item in (stdout, stderr)):
            return cmd

    return None


def is_conda_available() -> bool:
    """Check if conda is available in path."""
    return get_conda_cmd_path() is not None


def get_conda_info(allow_fixes: bool = False) -> typing.Optional['conda_types.CondaInfoOutput']:
    """Return conda info as a dictionary."""
    conda_cmd: typing.Optional[str] = get_conda_cmd_path()
    if conda_cmd is None:
        return None

    while True:
        out: str = launch.run_process(cmd_list=[conda_cmd, 'info', '--json']).stdout

        result: typing.Union['conda_types.CondaInfoOutput', 'conda_types.CondaErrorOutput']
        try:
            result = json.loads(out)
        except (TypeError, ValueError):
            return None

        if 'error' not in result:
            return typing.cast('conda_types.CondaInfoOutput', result)

        error: 'conda_types.CondaErrorOutput' = typing.cast('conda_types.CondaErrorOutput', result)
        if allow_fixes and (solvers.POOL.solve(error=error) is not None):
            continue
        raise exceptions.CondaError(error=error)



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name,missing-module-docstring

from qtpy import QtCore
from qtpy.QtWidgets import QTableView, QHeaderView  # pylint: disable=no-name-in-module
from qtpy import QtGui


class SelectableChannelsItem(QtGui.QStandardItem):  # pylint: disable=too-few-public-methods
    """
    Standard table item object representation.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setEditable(False)

    def __lt__(self, other):
        if self.isCheckable():
            return not self.checkState() in (other.checkState(), QtCore.Qt.Checked)
        return super().__lt__(other)


class SelectableChannelsListTable(QTableView):
    """
    Table object representation for selecting channels while doing login to Team Edition instance.
    """
    NAME_INDEX = 0
    MIRROR_INDEX = 1
    PRIVACY_INDEX = 2
    OWNERS_INDEX = 3
    DEFAULT_CHANNELS_INDEX = 4
    CHANNELS_INDEX = 5

    TABLE_HEADERS = ['Name', 'Mirror', 'Privacy', 'Owners', 'Add to default_channels', 'Add to channels']

    def __init__(self, parent, table_data, channels=None, default_channels=None):
        super().__init__()

        if not channels:
            channels = []
        if not default_channels:
            default_channels = []

        self.parent = parent
        self.apply_data_model(table_data, channels, default_channels)

        for index in range(len(self.TABLE_HEADERS)):
            self.horizontalHeader().setSectionResizeMode(index, QHeaderView.ResizeToContents)

        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().setVisible(False)
        self.setSortingEnabled(True)

        self.horizontalHeader().resizeSection(self.CHANNELS_INDEX, 120)

    def apply_data_model(self, table_data, channels, default_channels):
        """
        Creates the model data for the table depending on passed data in 'table_data' attribute.

        :param dict[str, str/int] table_data: The data which will be hold in model.
        :param list[str] channels: List of 'channels' names which must be checked by default.
        :param list[str] default_channels: List of 'default_channels' which must be checked by default.
        """
        model = QtGui.QStandardItemModel()
        model.setHorizontalHeaderLabels(self.TABLE_HEADERS)

        for row, channel in enumerate(table_data):
            model.setItem(row, self.NAME_INDEX, SelectableChannelsItem(channel['name']))
            model.setItem(row, self.MIRROR_INDEX, self.generate_checkable_item(bool(channel['mirror_count']), False))
            model.setItem(row, self.PRIVACY_INDEX, SelectableChannelsItem(channel['privacy']))
            model.setItem(row, self.OWNERS_INDEX, SelectableChannelsItem(', '.join(filter(bool, channel['owners']))))

            default_channels_is_checked = channel['name'] in default_channels
            channels_is_checked = channel['name'] in channels

            model.setItem(row, self.DEFAULT_CHANNELS_INDEX, self.generate_checkable_item(default_channels_is_checked))
            model.setItem(row, self.CHANNELS_INDEX, self.generate_checkable_item(channels_is_checked))

        self.setModel(model)
        self.sortByColumn(self.MIRROR_INDEX, QtCore.Qt.DescendingOrder)

    @staticmethod
    def generate_checkable_item(is_checked=False, is_enabled=True):
        """
        Creates a table item with a checkbox and empty string.

        :param bool is_checked: Defines if the checkbox must be checked.
        :param bool is_enabled: Defines if the checkbox must be enabled.

        :return SelectableChannelsItem: The created table item object.
        """
        checkable_item = SelectableChannelsItem()
        checkable_item.setCheckable(True)
        checkable_item.setEnabled(is_enabled)
        checkable_item.setTextAlignment(QtCore.Qt.AlignCenter)

        if is_checked:
            checkable_item.setCheckState(QtCore.Qt.Checked)

        return checkable_item

    def get_selected_channels(self):
        """
        Returns the lists with 'default_channels' and 'channels' names fetched from the model,
        where on row 4 and 5 (default_channels and channels) checkboxes are in the checked state.

        :return tuple[list[str], list[str]]
        """
        default_channels = []
        channels = []

        for row in range(self.model().rowCount()):
            data_index = self.model().index(row, self.NAME_INDEX)

            if self.model().item(row, self.DEFAULT_CHANNELS_INDEX).checkState() == QtCore.Qt.Checked:
                default_channels.append(self.model().data(data_index))

            if self.model().item(row, self.CHANNELS_INDEX).checkState() == QtCore.Qt.Checked:
                channels.append(self.model().data(data_index))

        return default_channels, channels



# -*- coding: utf-8 -*-

"""Common parts of different issue detectors and solvers."""

from __future__ import annotations

__all__ = ['SolverRecord', 'SolverCollection']

import typing

if typing.TYPE_CHECKING:
    import typing_extensions


TInvSolver = typing.TypeVar('TInvSolver')


class SolverRecord(typing.Generic[TInvSolver]):
    """Single registered solver."""

    __slots__ = ('__solver', '__tags')

    def __init__(self, solver: TInvSolver, tags: typing.Iterable[str] = ()) -> None:
        """Initialize new :class:`~SolverRecord` instance."""
        self.__solver: 'typing_extensions.Final[TInvSolver]' = solver
        self.__tags: 'typing_extensions.Final[typing.Set[str]]' = set(tags)

    @property
    def solver(self) -> TInvSolver:  # noqa: D401
        """Solver value."""
        return self.__solver

    @property
    def tags(self) -> typing.AbstractSet[str]:
        """Related tags to this solver."""
        return self.__tags

    def extend(self, *, tags: typing.Iterable[str] = ()) -> None:
        """Extend current record with additional tags."""
        self.__tags.update(tags)


class SolverCollection(typing.Generic[TInvSolver], typing.Iterable[SolverRecord[TInvSolver]], typing.Sized):
    """Collection of issue solvers."""

    __slots__ = ('__content', '__unique_tags')

    def __init__(self) -> None:
        """Initialize new :class:`~SolverCollection` instance."""
        self.__content: 'typing_extensions.Final[typing.List[SolverRecord[TInvSolver]]]' = []
        self.__unique_tags: 'typing_extensions.Final[typing.Dict[str, SolverRecord[TInvSolver]]]' = {}

    def only(
            self,
            *,
            tags: typing.Union[None, str, typing.Iterable[str]] = None,
    ) -> typing.Iterator[SolverRecord[TInvSolver]]:
        """Iterate through filtered subset of solvers."""
        # commentaries are the base to create a place for additional filters

        if tags is None:  # remove block
            yield from self.__content
            return

        if isinstance(tags, str):
            tags = {tags}
        else:  # elif tags is not None
            tags = set(tags)

        solver_record: SolverRecord[TInvSolver]
        for solver_record in self.__content:
            if not tags.intersection(solver_record.tags):  # if (tags is not None) and (...):
                continue

            yield solver_record

    @typing.overload
    def register(
            self,
            solver: TInvSolver,
            *,
            tags: typing.Union[str, typing.Iterable[str]] = (),
            unique_tags: typing.Union[str, typing.Iterable[str]] = (),
    ) -> TInvSolver:
        """Register a new unnamed solver."""

    @typing.overload
    def register(
            self,
            *,
            tags: typing.Union[str, typing.Iterable[str]] = (),
            unique_tags: typing.Union[str, typing.Iterable[str]] = (),
    ) -> typing.Callable[[TInvSolver], TInvSolver]:
        """Register a new named solver."""

    def register(self, solver=None, *, tags=(), unique_tags=()):
        """
        Register new solver.

        This method can be used as a decorator:

        .. code-block:: python

            solvers = SolverCollection()

            @solvers.register
            def unnamed_solver(context):
                ...

            @solvers.register(tags='some_tag', unique_tags=('unique_a', 'unique_b'))
            def named_solver(context):
                ...

        as well as direct function:

        .. code-block:: python

            def custom_solver(context):
                ...

            solvers.register(custom_solver, tags=('other_tag', 'additional_tag'), unique_tags='custom_solver')

        :param solver: Solver that should be registered.
        :param tags: Optional collection of common tags.
        :param unique_tags: Optional collection of tags, that should be unique only for this particular solver.
        """
        if isinstance(unique_tags, str):
            unique_tags = {unique_tags}
        else:
            unique_tags = set(unique_tags)

        if isinstance(tags, str):
            tags = {tags}
        else:
            tags = set(tags)
        tags.update(unique_tags)

        def wrapper(item: TInvSolver) -> TInvSolver:
            solver_record: typing.Optional[SolverRecord[TInvSolver]]
            for solver_record in self.__content:
                if solver_record.solver is item:
                    break
            else:
                solver_record = None

            common_unique_tags: typing.Set[str] = unique_tags.intersection(self.__unique_tags)
            if common_unique_tags:
                message: str = ', '.join(repr(item) for item in common_unique_tags)
                raise ValueError(f'some of the unique tags are already used: {message}')

            if solver_record is None:
                solver_record = SolverRecord(solver=item, tags=tags)
                self.__content.append(solver_record)
            else:
                solver_record.extend(tags=tags)

            tag: str
            for tag in unique_tags:
                self.__unique_tags[tag] = solver_record

            return item

        if solver is None:
            return wrapper
        return wrapper(solver)

    def __iter__(self) -> typing.Iterator[SolverRecord[TInvSolver]]:
        """Iterate through registered issue solvers."""
        return iter(self.__content)

    def __len__(self) -> int:
        """Get total number of registered solvers."""
        return len(self.__content)



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name,no-name-in-module,too-few-public-methods

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Widgets module."""

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtSvg
from qtpy import QtWidgets
from qtpy.QtCore import QSize, Qt, QUrl, Signal
from qtpy.QtGui import QDesktopServices, QPixmap
from qtpy.QtWidgets import (
    QCheckBox, QComboBox, QFrame, QLabel, QLineEdit, QMenu, QPushButton, QSizePolicy, QStackedWidget, QToolButton,
    QVBoxLayout, QWidget,
)
from anaconda_navigator.utils.qthelpers import add_actions, create_action, update_pointer


# --- Base widgets
# -----------------------------------------------------------------------------

class WidgetBase(QWidget):
    """Widget base implementation."""

    sig_hovered = Signal(bool)
    sig_focused = Signal(bool)

    def _fix_style(self):
        # Mac related issues
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)
        self.setFocusPolicy(Qt.StrongFocus)

    def _fix_layout(self, layout):
        if layout:
            layout.setSpacing(0)
            layout.setContentsMargins(0, 0, 0, 0)

            items = (layout.itemAt(i).widget() for i in range(layout.count()))
            for w in items:
                if w:
                    new_layout = w.layout()
                    self._fix_layout(new_layout)

    def focusInEvent(self, event):
        """Override Qt method."""
        QWidget.focusInEvent(self, event)
        self.setProperty('focused', True)

    def focusOutEvent(self, event):
        """Override Qt method."""
        QWidget.focusOutEvent(self, event)
        self.setProperty('focused', False)

    def enterEvent(self, event):
        """Override Qt method."""
        QWidget.enterEvent(self, event)
        self.setProperty('hovered', True)

    def leaveEvent(self, event):
        """Override Qt method."""
        QWidget.leaveEvent(self, event)
        self.setProperty('hovered', False)

    def setDisabled(self, value):
        """Override Qt method."""
        QWidget.setDisabled(self, value)
        self.setProperty('disabled', value)
        self.setProperty('enabled', not value)

    def setEnabled(self, value):
        """Override Qt method."""
        QWidget.setEnabled(self, value)
        self.setProperty('enabled', value)
        self.setProperty('disabled', not value)

    def setProperty(self, name, value):
        """Override Qt method."""
        QWidget.setProperty(self, name, value)
        self.style().unpolish(self)
        self.style().polish(self)
        self.update()

    def setLayout(self, layout):
        """Override Qt method."""
        self._fix_layout(layout)
        QWidget.setLayout(self, layout)


class ButtonBase(QPushButton, WidgetBase):
    """Base button used in CSS styling."""

    sig_entered = Signal()
    sig_left = Signal()

    def __init__(self, *args, **kwargs):
        """Base button used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setAutoDefault(False)
        self.setDefault(False)
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)
        self.setFocusPolicy(Qt.StrongFocus)

    def focusInEvent(self, event):
        """Override Qt method."""
        super().focusInEvent(event)
        self.sig_entered.emit()

    def focusOutEvent(self, event):
        """Override Qt method."""
        super().focusOutEvent(event)
        self.sig_left.emit()

    def enterEvent(self, event):
        """Override Qt method."""
        super().enterEvent(event)
        self.sig_entered.emit()

    def leaveEvent(self, event):
        """Override Qt method."""
        super().leaveEvent(event)
        self.sig_left.emit()

    def mousePressEvent(self, event):
        """Override Qt method."""
        super().mousePressEvent(event)
        self.setProperty('pressed', True)

    def mouseReleaseEvent(self, event):
        """Override Qt method."""
        super().mouseReleaseEvent(event)
        self.setProperty('pressed', False)


class ButtonToolBase(QToolButton, WidgetBase):
    """Base button used in CSS styling."""
    def __init__(self, parent=None, text=''):
        """Base button used in CSS styling."""
        super().__init__(parent=parent)
        self.setCheckable(False)
        self.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.setText(text)
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)


class CheckBoxBase(QCheckBox, WidgetBase):
    """Checkbox used in CSS styling."""

    def __init__(self, *args, **kwargs):
        """Checkbox used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX


class RadioButtonBase(QtWidgets.QRadioButton, WidgetBase):
    """RadioButton used in CSS styling."""

    def __init__(self, *args, **kwargs):
        """Checkbox used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX


class ComboBoxBase(QComboBox):
    """Combobox used in CSS styling."""
    def __init__(self, *args, **kwargs):
        """Combobox used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX

    def showPopup(self):
        """Override Qt method."""
        index = self.currentIndex()
        menu = QMenu(self)

        menu.setToolTipsVisible(True)
        actions = []

        for i in range(self.count()):
            tip = self.itemData(i, Qt.ToolTipRole)
            text = self.itemText(i)
            action = create_action(self, text, toggled=lambda v=None, i=i: self.setCurrentIndex(i), tip=tip)

            actions.append(action)
            action.setChecked(i == index)

        add_actions(menu, actions)
        menu.setFixedWidth(self.width())
        bottom_left = self.contentsRect().bottomLeft()
        menu.popup(self.mapToGlobal(bottom_left))


class FrameBase(QFrame, WidgetBase):
    """Button used in CSS styling."""
    def __init__(self, *args, **kwargs):
        """Button used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(Qt.NoFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX

    def setLayout(self, layout):
        """Override Qt method."""
        self._fix_layout(layout)
        super().setLayout(layout)


class LabelBase(QLabel, WidgetBase):
    """Label used in CSS styling."""

    sig_clicked = Signal()

    def __init__(self, *args, **kwargs):
        """Label used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(Qt.NoFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX

    def mousePressEvent(self, ev: QtGui.QMouseEvent) -> None:
        """
        Process mouse pressing event.

        Used to detect clicking on enabled label.
        """
        if self.isEnabled() and (ev.button() == QtCore.Qt.LeftButton):
            self.sig_clicked.emit()


class StackBody(QStackedWidget, WidgetBase):
    """Stacked widget used in CSS styling of main custom bar stack."""
    def __init__(self, *args, **kwargs):
        """Stacked widget used in CSS styling of main custom bar stack."""
        super().__init__(*args, **kwargs)
        self.setContentsMargins(0, 0, 0, 0)
        self.setFrameStyle(QFrame.NoFrame)
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX

    def setLayout(self, layout):
        """Override Qt method."""
        self._fix_layout(layout)
        super().setLayout(layout)


class LineEditBase(QLineEdit):
    """Line edit used in CSS styling."""
    def __init__(self, *args, **kwargs):
        """Line edit used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX

    def mousePressEvent(self, event):
        """Override Qt method."""
        super().mousePressEvent(event)
        self.setProperty('pressed', True)

    def mouseReleaseEvent(self, event):
        """Override Qt method."""
        super().mouseReleaseEvent(event)
        self.setProperty('pressed', False)


# --- Buttons
# -----------------------------------------------------------------------------
class ButtonToolNormal(ButtonToolBase):
    """Button used in CSS styling."""


class ButtonNormal(ButtonBase):
    """Button used in CSS styling."""


class ButtonPrimary(ButtonBase):
    """Button used in CSS styling."""


class ButtonDanger(ButtonBase):
    """Button used in CSS styling."""


class ButtonLink(QPushButton, WidgetBase):
    """
    Button use to represent a clickable (and keyboard focusable) web link.

    It is styled to be used as a label.
    """

    sig_hovered = Signal(bool)

    def __init__(self, *args, **kwargs):
        """
        Button use to represent a clickable (and keyboard focusable) web link.

        It is styled to be used as a label.
        """
        super().__init__(*args, **kwargs)
        self.setAutoDefault(False)
        self.setDefault(False)
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX

    def enterEvent(self, event):
        """Override Qt method."""
        super().enterEvent(event)
        update_pointer(Qt.PointingHandCursor)
        self.sig_hovered.emit(True)

    def leaveEvent(self, event):
        """Override Qt method."""
        super().leaveEvent(event)
        update_pointer()
        self.sig_hovered.emit(False)


class ButtonLabel(QPushButton):
    """
    A button that is used next to ButtonLink to avoid missalignments.

    It looks and acts like a label.
    """
    def __init__(self, *args, **kwargs):
        """A button that is used next to ButtonLink to avoid missalignments."""
        super().__init__(*args, **kwargs)
        self.setDisabled(True)
        self.setFocusPolicy(Qt.NoFocus)
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on mac


# --- Frames
# -----------------------------------------------------------------------------
class FrameTabBar(FrameBase):
    """Frame used in CSS styling."""


class FrameTabBody(FrameBase):
    """Frame used in CSS styling."""


class FrameTabHeader(FrameBase):
    """Frame used in CSS styling."""


class FrameTabContent(FrameBase):
    """Frame used in CSS styling."""


class FrameTabFooter(FrameBase):
    """Frame used in CSS styling."""


# --- Labels
# -----------------------------------------------------------------------------
class LabelLinks(LabelBase):
    """Label link used as url link."""

    LINKS_STYLE = """<style>
    a {
        color:green;
        text-decoration: underline;
    }
    </style>
    """

    def __init__(self, *args, **kwargs):
        """Label link used as url link."""
        super().__init__(*args, **kwargs)
        self.setOpenExternalLinks(False)
        self.linkActivated.connect(self._link_activated)
        self._original_text = self.text()
        self._add_style()

    def _add_style(self):
        text = self._original_text
        if self.LINKS_STYLE not in text:
            self.setText(self.LINKS_STYLE + text)

    @staticmethod
    def _link_activated(url):
        QDesktopServices.openUrl(QUrl(url))
        from anaconda_navigator.utils.analytics import GATracker  # pylint: disable=import-outside-toplevel
        tracker = GATracker()
        tracker.track_event('content', 'link', url)

    def setText(self, text):
        """Override Qt method."""
        self._original_text = text
        super().setText(text)
        self._add_style()


class LabelImageLinkVertical(QWidget):  # pylint: disable=missing-class-docstring
    def __init__(self, text, url, image, width, height, margin):  # pylint: disable=too-many-arguments
        """
        Widget which can be used to show advertisements or other data by output
        an image with the text.

        Clicking on it will redirect to URL which is setup in the 'url' param.

        :param str text: The text to be displayed.
        :param url text: The url to redirect on the 'on click' event.
        :param bytes image: The image to be displayed.
        :param int width: The width of the image.
        :param int height: The height of the image.
        :param int margin: The margin to be used for the widget.
        """
        super().__init__()
        self.setFocusPolicy(Qt.StrongFocus)  # Needed on OSX
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)  # Needed on OSX

        # Data to display.
        self.text = text
        self.url = url
        self.image = image

        # Setup elements to display.
        self.vbox = QVBoxLayout()

        if image:
            pixmap = QPixmap()
            pixmap.loadFromData(self.image)
            scaled_pixmap = pixmap.scaled(width, height, Qt.KeepAspectRatio)

            self.image_label = QLabel()
            self.image_label.setScaledContents(True)
            self.image_label.setPixmap(scaled_pixmap)
            self.image_label.setContentsMargins(margin, margin, margin, margin)
            self.image_label.setFixedSize(width + margin, height + margin)
            self.vbox.addWidget(self.image_label, alignment=Qt.AlignCenter)
            self.vbox.addSpacing(5)
            if not text:
                self.vbox.addSpacing(5)

        # maximum 60 symbols
        if text:
            self.text_label = QLabel()
            self.text_label.setText(self.text)
            self.text_label.setWordWrap(True)
            self.text_label.setAlignment(Qt.AlignCenter)
            self.text_label.setContentsMargins(margin, margin, margin, margin)
            self.text_label.setFixedSize(width + margin, 50)  # 2 lines
            self.vbox.addWidget(self.text_label, alignment=Qt.AlignCenter)
            self.vbox.addSpacing(5)

        self.setLayout(self.vbox)

    def enterEvent(self, event):
        """
        Default PyQt enterEvent override. Changes the text color
        on the current widget when the mouse is entered the widget.
        """
        super().enterEvent(event)
        update_pointer(Qt.PointingHandCursor)

    def leaveEvent(self, event):
        """
        Default PyQt leaveEvent override. Changes the text color to default
        on the current widget when the mouse is out of the widget
        """
        super().leaveEvent(event)
        update_pointer()

    def mousePressEvent(self, QMouseEvent):  # pylint: disable=unused-argument
        """
        Default PyQt mousePressEvent override. Opens a new tab in the default
        web browser with the url defined in the 'url' attribute.
        Also does analytics tracking actions.
        """
        QDesktopServices.openUrl(QUrl(self.url))
        from anaconda_navigator.utils.analytics import GATracker  # pylint: disable=import-outside-toplevel
        tracker = GATracker()
        tracker.track_event('content', 'link', self.url)


# --- Spacers
# -----------------------------------------------------------------------------
class SpacerHorizontal(LabelBase):
    """Label used in CSS styling."""


class SpacerVertical(LabelBase):
    """Label used in CSS styling."""


# --- Other Buttons
# -----------------------------------------------------------------------------
class ButtonPrimaryAction(QPushButton):
    """Button used in CSS styling."""


class ButtonCancel(QPushButton):
    """Button used in CSS styling."""


class ButtonSecondaryTextual(QPushButton):
    """Button used in CSS styling."""


class ButtonSecondaryIcon(QPushButton):
    """Button used in CSS styling."""


class ButtonEnvironmentCancel(QPushButton):
    """Button used in CSS styling."""


class ButtonEnvironmentPrimary(QPushButton):
    """Button used in CSS styling."""


# --- Buttons that are used as labels.
class ButtonEnvironmentOptions(ButtonBase):
    """Button used in CSS styling."""


class FrameEnvironments(QFrame):
    """Frame used in CSS styling."""


class FrameEnvironmentsList(QFrame):
    """Frame used in CSS styling."""


class FrameEnvironmentsListButtons(QFrame):
    """Frame used in CSS styling."""


class FrameEnvironmentsPackages(QFrame):
    """Frame used in CSS styling."""


class QSvgWidget(QtSvg.QSvgWidget):
    """SvgWidget that knows about its size ratio."""
    def __init__(self, *args, **kwargs):
        """SvgWidget that knows about its size ratio."""
        super().__init__(*args, **kwargs)

        # Variables
        self._path = None
        if args:
            self._path = args[0]
        self._default_size = self._get_default_size()

        # Widget setup
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)

    def _get_default_size(self):
        """Get dfault size of SVG image."""
        size = QSize()
        if self._path:
            item = QtSvg.QGraphicsSvgItem(self._path)
            size = item.renderer().defaultSize()
        return size

    def size_for_width(self, width):
        """Return the size for a given width, preserving the ratio."""
        size = self.default_size()
        ratio_w_h = size.width() / size.height()
        return QSize(width, width / ratio_w_h)

    def size_for_height(self, height):
        """Return the size for a given height, preserving the ratio."""
        size = self.default_size()
        ratio_w_h = size.width() / size.height()
        return QSize(height * ratio_w_h, height)

    def default_size(self):
        """Return the default size of the SVG image."""
        return self._default_size
